# -*- coding: utf-8 -*-
'''
Created on 28.10.2012

@author: dominik
'''
import logging

from mock import MagicMock, Mock
from unittest import TestCase

import quidditas

##############################################
###               Test Data                ###
##############################################

class TestingComponentA(object):
    updated = 0

class TestingComponentB(object):
    updated = 0

class TestingComponentC(object):
    pass

class TestingProcessorA(quidditas.Processor):
    component_types = (TestingComponentA, )

    def update_entity(self, time, entity, components):
        pass

class TestingProcessorB(quidditas.Processor):
    component_types = (TestingComponentB, TestingComponentA)

class TestingProcessorC(quidditas.Processor):
    pass

##############################################
###              Test Cases                ###
##############################################

class TestQuidditas(TestCase):
    """ Tests the quidditas class. """
    def setUp(self):
        self.q = quidditas.Quidditas()
        self.q.entity_manager = MagicMock()
        self.q.component_manager = MagicMock()
        self.q.processor_manager = MagicMock()

    def test_quidditas_init(self):
        """ Test if all managers are created and have
        the necessary back linkgs. """
        self.q = quidditas.Quidditas()
        assert self.q.entity_manager is not None
        assert self.q.component_manager is not None
        assert self.q.processor_manager is not None
        assert self.q.entity_manager.quidditas is not None
        assert self.q.component_manager.quidditas is not None
        assert self.q.processor_manager.quidditas is not None

    def test_manual_add_entity(self):
        """ Test if an entity that is manually added to the
        world gets accepted into the system and receives
        an id. """
        e = quidditas.Entity()
        self.q.add_entity(e)

        assert e.quidditas is self.q
        self.q.entity_manager.add_entity.assert_called_once_with(e)
        self.q.component_manager.add_components.is_called_once()

    def test_add_entity_by_type(self):
        """ DOCME """
        comp = TestingComponentA()
        entity = Mock()
        entity.components = (comp,)
        self.q.entity_manager = Mock()
        self.q.component_manager = Mock()
        self.q.processor_manager = Mock()
        self.q.entity_manager.create_entity.return_value = entity

        self.q.create_entity("test_type")
        self.q.entity_manager.create_entity.assert_called_with("test_type")
        assert entity.quidditas is self.q
        self.q.component_manager.add_components.assert_called_with(entity.gid, (comp,))
        self.q.processor_manager.refresh_entity.assert_called_with(entity)

    def test_update_call_to_process_manager(self):
        """ Test that the update call is forwarded to the process
        manager. """
        self.q.update(1)
        self.q.processor_manager.update.assert_called_once_with(1)

    def test_add_remove_processor(self):
        """ DOCME """
        proc = TestingProcessorA()
        proc.refresh_entities = MagicMock()
        self.q.processor_manager.add_processor.return_value = True
        entity_list = (quidditas.Entity(),)
        self.q.entity_manager.entities.values.return_value = entity_list

        self.q.add_processor(proc)

        self.q.processor_manager.add_processor.assert_called_once_with(proc)
        assert proc.component_manager is self.q.component_manager
        proc.refresh_entities.assert_called_once_with(entity_list)

    def test_remove_processor(self):
        """ DOCME """
        proc = TestingProcessorA()
        self.q.remove_processor(proc)
        self.q.processor_manager.remove_processor.called_once_with(proc)

    def test_remove_entity_gid(self):
        """ DOCME """
        e = quidditas.Entity()
        self.q.remove_entity_gid(e.gid)

        assert not self.q.entity_manager.remove_entity_gid.called
        assert not self.q.component_manager.remove_entity_gid.called
        assert not self.q.processor_manager.refresh_entity_gid.called

        self.q.update(1)

        self.q.entity_manager.remove_entities_gid.assert_called_once()
        self.q.component_manager.remove_entities_gid.assert_called_once()
        self.q.processor_manager.refresh_entities_gid.assert_called_once()

    def test_add_component(self):
        """ DOCME """
        entity = quidditas.Entity()
        comp = TestingComponentA()
        self.q.add_component(entity, comp)
        self.q.component_manager.add_component.assert_called_once_with(entity.gid, comp)
        self.q.processor_manager.refresh_entity.assert_called_once_with(entity)

    def test_remove_component(self):
        """ DOCME """
        entity = quidditas.Entity()
        comp = TestingComponentA()
        self.q.remove_component(entity, comp)
        self.q.component_manager.remove_component.assert_called_once_with(entity.gid, comp.__class__)
        self.q.processor_manager.refresh_entity.assert_called_once_with(entity)

class TestEntity(TestCase):
    """ Test the behaviour of the entity's functions. """
    def setUp(self):
        self.q = MagicMock()
        self.e = quidditas.Entity()

    def test_kill(self):
        """ Test that the entity is removed from the system on kill. """
        self.e.quidditas = self.q
        self.e.kill()
        self.q.remove_entity_gid.assert_called_once()

    def test_add_component_with_quidditas(self):
        """ Test adding a component to an entity that is registered with a quidditas instance. """
        comp = TestingComponentA()
        self.e.quidditas = self.q
        self.e.add_component(comp)
        with self.assertRaises(AttributeError):
            comp in self.e.components
        self.q.add_component.assert_called_once_with(self.e, comp)

    def test_add_component_without_quidditas(self):
        """ Test entity behaviour when adding components before
        adding the entity to a quidditas instance. """
        comp = TestingComponentA()
        self.e.add_component(comp)
        assert not self.q.add_component.called, "Adding component should not have been forwarded."

    def test_remove_component_with_quidditas(self):
        """ Test the entity behaviour when adding and removing components
        to and from the entity before it is added to a quidditas instance. """
        self.q.remove_component = MagicMock()
        compA = TestingComponentA()
        compB = TestingComponentB()
        self.q.add_entity(self.e)
        self.e.add_component(compA)
        self.e.add_component(compB)
        self.e.remove_component(compA)
        with self.assertRaises(AttributeError):
            compA in self.e.components

        self.q.remove_component.assert_called_once_with(self.e, compA)

    def test_get_component_without_quidditas(self):
        """ DOCME """
        comp = TestingComponentA()
        self.e.remove_component(comp) #Nothing happens

        with self.assertRaises(LookupError):
            # Exception as no component for this type is available.
            self.e.get_component(TestingComponentA)

        self.e.add_component(comp)
        assert self.e.get_component(TestingComponentA) is comp
        assert self.e.get_component(TestingComponentB) is None

    def test_get_component_with_quidditas(self):
        """ DOCME """
        comp = TestingComponentA()
        self.q = quidditas.Quidditas()
        self.q.add_entity(self.e)
        self.e.add_component(comp)
        assert self.e.get_component(TestingComponentA) is comp
        with self.assertRaises(LookupError):
            self.e.get_component(TestingComponentB)

    def test_remove_component_with_quidditas(self):
        """ Test the component removal on an entity that is added to quidditas."""
        compA = TestingComponentA()
        compB = TestingComponentB()
        self.e.add_component(compA)
        self.e.add_component(compB)
        self.e.remove_component(compA)
        assert compA not in self.e.components
        assert compB in self.e.components

    def test_get_component_types_with_quidditas(self):
        """ DOCME """
        compA = TestingComponentA()
        compB = TestingComponentB()
        self.q = quidditas.Quidditas()

        self.q.add_entity(self.e)
        assert self.e.quidditas is self.q

        assert len(self.e.get_component_types()) == 0

        self.e.add_component(compA)
        assert TestingComponentA in self.e.get_component_types()
        assert TestingComponentB not in self.e.get_component_types()

        self.e.add_component(compB)
        assert TestingComponentA in self.e.get_component_types()
        assert TestingComponentB in self.e.get_component_types()

        self.e.remove_component(compA)
        assert TestingComponentA not in self.e.get_component_types()
        assert TestingComponentB in self.e.get_component_types()

    def test_get_component_types_without_quidditas(self):
        """ DOCME """
        compA = TestingComponentA()
        compB = TestingComponentB()

        assert len(self.e.get_component_types()) == 0

        self.e.add_component(compA)
        assert TestingComponentA in self.e.get_component_types()
        assert TestingComponentB not in self.e.get_component_types()

        self.e.add_component(compB)
        assert TestingComponentA in self.e.get_component_types()
        assert TestingComponentB in self.e.get_component_types()

        self.e.remove_component(compA)
        assert TestingComponentA not in self.e.get_component_types()
        assert TestingComponentB in self.e.get_component_types()

class TestComponentManager(TestCase):
    """ DOCME """
    def setUp(self):
        self.q = quidditas.Quidditas()
        self.comp_manager = self.q.component_manager

    def test_add_component(self):
        """ DOCME """
        e = quidditas.Entity()
        comp = TestingComponentA()
        comp2 = TestingComponentA()

        assert not self.comp_manager.entity_components.has_key(e.gid)

        self.comp_manager.add_component(e.gid, comp)
        assert e.gid in self.comp_manager.entity_components.keys()

        components = self.comp_manager.entity_components[e.gid]
        assert comp in components.values()
        assert len(components.values()) == 1

        self.comp_manager.add_component(e.gid, comp2)
        assert comp not in components.values()
        assert comp2 in components.values()
        assert len(components.values()) == 1


    def test_remove_component(self):
        """ DOCME """
        e = quidditas.Entity()
        compA = TestingComponentA()
        compB = TestingComponentB()

        self.comp_manager.add_component(e.gid, compA)
        self.comp_manager.add_component(e.gid, compB)
        assert len(self.comp_manager.entity_components[e.gid].values()) == 2

        self.comp_manager.remove_component(e.gid, compA.__class__)
        components = self.comp_manager.entity_components[e.gid]
        assert compA not in components.values()
        assert compB in components.values()
        self.comp_manager.remove_component(e.gid, compB.__class__)
        self.comp_manager.remove_component(e.gid, compA.__class__) # Nothing happens
        assert compB not in components.values()

    def test_add_components(self):
        """ DOCME """
        entity = quidditas.Entity()
        compA = TestingComponentA()
        compA2 = TestingComponentA()
        compB = TestingComponentB()
        compC = TestingComponentC()

        self.comp_manager.add_components(entity.gid, (compA, compB))
        assert compA in self.comp_manager.entity_components[entity.gid].values()
        assert compB in self.comp_manager.entity_components[entity.gid].values()

        # Merging new components with old
        self.comp_manager.add_components(entity.gid, (compC,))
        assert compA in self.comp_manager.entity_components[entity.gid].values()
        assert compB in self.comp_manager.entity_components[entity.gid].values()
        assert compC in self.comp_manager.entity_components[entity.gid].values()

        # Replacing components
        self.comp_manager.add_components(entity.gid, (compA2,))
        assert compA2 in self.comp_manager.entity_components[entity.gid].values(), "compA2 should have replaced compA"
        assert compA not in self.comp_manager.entity_components[entity.gid].values()

    def test_remove_entity_gid_non_existent(self):
        """ DOCME """
        e = quidditas.Entity()
        self.comp_manager.remove_entity_gid(e.gid)
        # No exception thrown, nothing happens

    def test_remove_entity_gid(self):
        """ DOCME """
        e = quidditas.Entity()
        self.comp_manager.add_component(e.gid, TestingComponentA())
        assert e.gid in self.comp_manager.entity_components.keys()
        self.comp_manager.remove_entity_gid(e.gid)
        assert e.gid not in self.comp_manager.entity_components.keys()

    def test_remove_entities_gid(self):
        """ DOCME """
        e1 = quidditas.Entity()
        e1.gid = 0
        e2 = quidditas.Entity()
        e2.gid = 1

        self.comp_manager.add_component(e1.gid, TestingComponentA())
        self.comp_manager.add_component(e2.gid, TestingComponentA())
        assert e1.gid in self.comp_manager.entity_components.keys()
        assert e2.gid in self.comp_manager.entity_components.keys()

        self.comp_manager.remove_entities_gid((e1.gid, e2.gid))
        assert e1.gid not in self.comp_manager.entity_components.keys()
        assert e2.gid not in self.comp_manager.entity_components.keys()

class TestProcessorManager(TestCase):
    """ Test functionality of ProcessManager. """
    def setUp(self):
        self.q = quidditas.Quidditas()
        self.proc_manager = self.q.processor_manager

    def test_add_processors(self):
        """ DOCME """
        procA = TestingProcessorA()
        procB = TestingProcessorB()
        assert len(self.proc_manager.processors) == 0

        self.proc_manager.add_processor(procA)
        self.proc_manager.add_processor(procB)
        assert procA in self.proc_manager.processors
        assert procB in self.proc_manager.processors
        assert TestingProcessorA in self.proc_manager.processor_types
        assert TestingProcessorB in self.proc_manager.processor_types

    def test_remove_processor(self):
        """ DOCME """
        procA = TestingProcessorA()
        procB = TestingProcessorB()
        self.proc_manager.add_processor(procA)
        self.proc_manager.add_processor(procB)

        self.proc_manager.remove_processor(procB)
        assert procA in self.proc_manager.processors
        assert procB not in self.proc_manager.processors
        self.proc_manager.remove_processor(procB) # Nothing happens
        self.proc_manager.remove_processor(procA)
        assert procA not in self.proc_manager.processors

    def test_update(self):
        """ Test that processors are not automatically updated. """
        procA = TestingProcessorA()
        procB = TestingProcessorB()
        self.proc_manager.add_processor(procA)
        self.proc_manager.add_processor(procB)

        procA.update = MagicMock()
        procB.update = MagicMock()

        self.proc_manager.update(23)

        assert not procA.update.called
        assert not procB.update.called

    def test_refresh_entity(self):
        """ Test that refresh calls are forwarded to all
        assigned processors. """
        procA = TestingProcessorA()
        procB = TestingProcessorB()
        self.proc_manager.add_processor(procA)
        self.proc_manager.add_processor(procB)

        procA.refresh_entity = MagicMock()
        procB.refresh_entity = MagicMock()
        e = quidditas.Entity()

        self.proc_manager.refresh_entity(e)
        procA.refresh_entity.assert_called_once_with(e)
        procB.refresh_entity.assert_called_once_with(e)

        ### Remove one processor and call again.
        procA.refresh_entity = MagicMock()
        procB.refresh_entity = MagicMock()
        self.proc_manager.remove_processor(procB)
        self.proc_manager.refresh_entity(e)
        procA.refresh_entity.assert_called_once_with(e)
        assert not procB.refresh_entity.called

    def test_remove_entities_gid(self):
        """ DOCME """
        procA = MagicMock()
        self.proc_manager.add_processor(procA)
        self.proc_manager.remove_entities_gid(set((1,2,3)))
        procA.remove_entities_gid.assert_called_once_with(set((1,2,3)))

    def test_no_doubled_processor_types(self):
        """ DOCME """
        procA = TestingProcessorA()
        procB = TestingProcessorA()

        self.proc_manager.add_processor(procA)
        self.proc_manager.add_processor(procB)
        assert procA in self.proc_manager.processors
        assert procB not in self.proc_manager.processors
        assert procB.component_manager is None

class TestEntityManager(TestCase):
    """ DOCME """
    def setUp(self):
        self.q = quidditas.Quidditas()
        self.entity_manager = self.q.entity_manager
        self.types = {"typeA" : (TestingComponentA(),),
                      "typeAB" : (TestingComponentA(), TestingComponentB())}

    def test_merging_type_definitions(self):
        """ DOCME """
        self.entity_manager.add_type_definitions(self.types)
        assert "typeB" not in self.entity_manager.types.keys()

        self.entity_manager.add_type_definitions({"typeB":(TestingComponentB(),)})
        assert "typeB" in self.entity_manager.types.keys()
        assert "typeA" in self.entity_manager.types.keys()
        assert "typeAB" in self.entity_manager.types.keys()

    def test_create_entity_by_type(self):
        """ DOCME """
        self.entity_manager.add_type_definitions(self.types)
        with self.assertRaises(KeyError):
            entity = self.entity_manager.create_entity("does_not_exist")

        entityA = self.entity_manager.create_entity("typeA")
        assert entityA is not None
        assert entityA.get_type() == "typeA"
        assert TestingComponentA in entityA.get_component_types()
        assert TestingComponentB not in entityA.get_component_types()

        entityAB = self.entity_manager.create_entity("typeAB")
        assert entityAB.get_type() != "typeA"
        assert entityAB.get_type() == "typeAB"
        assert entityAB is not None, "TypeAB not instantiated."
        assert entityA.get_component(TestingComponentA) is not entityAB.get_component(TestingComponentA)

    def test_remove_entity_gid(self):
        """ DOCME """
        e = quidditas.Entity()
        with self.assertRaises(KeyError):
            self.entity_manager.remove_entity_gid(e.gid)
        self.entity_manager.add_entity(e)
        assert e.gid in self.entity_manager.entities.keys()
        self.entity_manager.remove_entity_gid(e.gid)
        assert e not in self.entity_manager.entities

    def test_remove_entities_gid(self):
        """ DOCME """
        e1 = quidditas.Entity()
        e1.gid = 0
        e2 = quidditas.Entity()
        e2.gid = 1
        self.entity_manager.add_entity(e1)
        self.entity_manager.add_entity(e2)
        assert e1.gid in self.entity_manager.entities.keys()
        assert e2.gid in self.entity_manager.entities.keys()

        self.entity_manager.remove_entities_gid(set((e1.gid, e2.gid)))
        assert e1.gid not in self.entity_manager.entities.keys()
        assert e2.gid not in self.entity_manager.entities.keys()

class TestProcessor(TestCase):
    """ DOCME """
    def setUp(self):
        self.q = quidditas.Quidditas()
        self.procA = TestingProcessorA()
        self.procB = TestingProcessorB()

    def test_init_without_components(self):
        """ Test behaviour of a processor that has no component types associated.
        Processors without component types should be possible as updateable processors
        that don't work on entities but on something else? """
        procC = TestingProcessorC()
        assert len(procC.component_types) == 0

    def test_refresh_entity(self):
        e = quidditas.Entity()
        comp = TestingComponentA()
        e.add_component(comp)

        self.procA.refresh_entity(e)
        logging.debug("Entities of procA: {}".format(self.procA.entities))
        assert e in self.procA.entities

        self.procB.refresh_entity(e)
        logging.debug("Entities of procB: {}".format(self.procB.entities))
        assert e not in self.procB.entities

        e.remove_component(comp)
        self.procA.refresh_entity(e)
        logging.debug("Entities of procA: {}".format(self.procA.entities))
        assert e not in self.procA.entities
        self.procB.refresh_entity(e)
        logging.debug("Entities of procB: {}".format(self.procB.entities))
        assert e not in self.procB.entities

    def test_update(self):
        """ DOCME """
        e = quidditas.Entity()
        self.q.add_entity(e)
        self.q.add_processor(self.procA)
        comp = TestingComponentA()
        e.add_component(comp)
        self.procA.refresh_entity(e)
        self.procA.update_entity = MagicMock()
        self.procA.update(1)
        self.procA.update_entity.assert_called_once_with(1, e, {TestingComponentA:comp})

class TestProcessor(TestCase):
    """ DOCME """
    def setUp(self):
        self.q = quidditas.Quidditas()

    def test_remove_entities_gid(self):
        """ DOCME """
        e1 = quidditas.Entity()
        e1.gid = 0
        e2 = quidditas.Entity()
        e2.gid = 1
        proc = TestingProcessorA()
        proc.entities.add(e1)
        proc.entities.add(e2)
        assert e1 in proc.entities
        assert e2 in proc.entities

        proc.remove_entities_gid(set((e1.gid, e2.gid, 99)))
        assert e1 not in proc.entities
        assert e2 not in proc.entities



"""
Copyright (c) 2012, Dominik Schacht
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
"""
import copy

class Quidditas(object):
    """ Provide a usable interface for the Quidditas system.
    """

    def __init__(self):
        """ Initiate an instance of the Quidditas system. """
        self.entity_manager = EntityManager(self)
        self.component_manager = ComponentManager(self)
        self.processor_manager = ProcessorManager(self)

        self.add_type_definitions = self.entity_manager.add_type_definitions

        self.entities_to_remove = set()
        self.entities_to_refresh = set()

    def get_entity(self, gid):
        """ Return an entity with the given gid. """
        return self.entity_manager.entities[gid]

    def add_processor(self, processor):
        """ Add the processor to the entity system.
        Return the processor when it was added, None when not. """
        if self.processor_manager.add_processor(processor):
            processor.set_component_manager(self.component_manager)
            processor.set_quidditas(self)
            processor.refresh_entities(self.entity_manager.entities.values())
            return processor
        return None

    def remove_processor(self, processor):
        """ Remove the processor from the entity system. """
        self.processor_manager.remove_processor(processor)

    def create_entity(self, type_name):
        """ Create an entity by its type name.
        Forwarded to the entity manager. """
        entity = self.entity_manager.create_entity(type_name)
        entity.quidditas = self
        self.component_manager.add_components(entity.gid, entity.components)
        self.processor_manager.refresh_entity(entity)
        del(entity.components)
        return entity

    def add_entity(self, entity):
        """ Add a manually created entity to the system.
        Forwarded to the entity manager. """
        self.entity_manager.add_entity(entity)
        entity.quidditas = self

        try:
            self.component_manager.add_components(entity.gid, entity.components)
        except AttributeError:
            # Entity did not have any components cached. Do nothing
            pass

        self.processor_manager.refresh_entity(entity)
        return entity

    def remove_entity_gid(self, gid):
        """ Schedule an entity for removal on next update. """
        if gid not in self.entities_to_remove:
            self.entities_to_remove.add(gid)

    def add_component(self, entity, component):
        """ Add a component to an entity. """
        self.component_manager.add_component(entity.gid, component)
        self.processor_manager.refresh_entity(entity)

    def remove_component(self, entity, component):
        self.component_manager.remove_component(entity.gid, component.__class__)
        self.processor_manager.refresh_entity(entity)

    def update(self, time):
        """ Perform an update step on all processors.
        time -- Time that has passed since last update. """
        if self.entities_to_remove:
            # Notify managers over removed entities.
            self.entity_manager.remove_entities_gid(self.entities_to_remove)
            self.component_manager.remove_entities_gid(self.entities_to_remove)
            self.processor_manager.remove_entities_gid(self.entities_to_remove)
            self.entities_to_remove.clear()
        if self.entities_to_refresh:
            self.entities_to_refresh.clear()

        self.processor_manager.update(time)

class EntityManager(object):
    """ Adds and removes entities from the world, notifies Processors and keeps
    them updated.

    Attributes:
    last_gid -- Id for the next entity that is created or added.
    entities -- Maps ids to entities for all existing entities.
    types -- Maps type names to a list of configured component instances, therefore it
             maps a type name to the prototype of that entity.
    """

    def __init__(self, quidditas):
        """ Create an empty entity manager. """
        self.next_gid = 0
        self.entities = {}
        self.types = {}
        self.quidditas = quidditas

    def create_entity(self, type_name):
        """ Create an entity of the specified type. """
        # Create an entity and add a copy of every component specified in the prototype.
        entity = Entity()
        # TODO: Remove logging
        print "Created entity of type '{}'.".format(type_name)
        entity.components = [copy.copy(comp) for comp in self.types[type_name]]
        entity.components.append(TypeComponent(type_name))
        return self.add_entity(entity)

    def add_entity(self, entity):
        """ Add a manually created entity to the world. """
        entity.gid = self.next_gid
        self.next_gid += 1
        self.entities[entity.gid] = entity
        return entity

    def remove_entities_gid(self, gids):
        """ Remove all entities specified in the list of gids. """
        [self.entities.pop(gid) for gid in gids]

    def remove_entity_gid(self, gid):
        """ Remove an entity from the world by its id. """
        return self.entities.pop(gid)

    def add_type_definitions(self, definitions):
        """ Merge the type definitions with the current type definitions. """
        self.types = dict(self.types.items() + definitions.items())

class ComponentManager(object):
    """ Manages the components and their associations with the existing entities.
    DOCME: attributes and maps """

    def __init__(self, quidditas):
        # {entity_gid => {component_type=>component_instance,..}, ...}
        self.entity_components = {}
        self.component_to_entities = {}
        self.quidditas = quidditas

    def get_entities_with_component(self, component_type):
        """ Return a list of entity-ids that have the given component type associated with them. """
        try:
            return self.component_to_entities[component_type]
        except KeyError:
            return None

    def add_component(self, entity_gid, component):
        """ Add the component to the id->components map. """
        try:
            components = self.entity_components[entity_gid]
            if components.has_key(component.__class__):
                # Component already assigned to this entity. Do something before overwrite?
                pass
            components[component.__class__] = component

        except KeyError:
            # No components assigned yet.
            self.entity_components[entity_gid] = {component.__class__ : component}
        try:
            self.component_to_entities[component.__class__].add(entity_gid)
        except KeyError:
            self.component_to_entities[component.__class__] = set((entity_gid,))

    def add_components(self, entity_gid, new_components):
        """ Add multiple components. """
        try:
            components = self.entity_components[entity_gid]
            for component in new_components:
                if components.has_key(component.__class__):
                    # Component already assigned to this entity. Do something before overwrite?
                    pass
                components[component.__class__] = component
        except KeyError:
            # No components assigned yet.
            self.entity_components[entity_gid] = {comp.__class__: comp for comp in new_components }

        # Add connections to component:[entity_gids] map
        for component in new_components:
            try:
                self.component_to_entities[component.__class__].add(entity_gid)
            except KeyError:
                self.component_to_entities[component.__class__] = set((entity_gid,))

    def remove_component(self, entity_gid, component_type):
        """ Remove a component from an entity. """
        try:
            # FIXME: This is wrong, see the attribute documentation for self.entity_components!
            components = self.entity_components[entity_gid]
            del(components[component_type])
        except KeyError:
            # Invalid entity id
            pass
        try:
            self.component_to_entities[component_type].remove(entity_gid)
        except KeyError:
            # Invalid entity id
            pass

    def remove_entity_gid(self, gid):
        """ React that an entity has been removed. """
        try:
            # remove entity from all component:[entity_gids] lists
            for component in self.entity_components[gid].keys():
                self.component_to_entities[component].remove(gid)

            # then remove it from the entity:[components] map
            del self.entity_components[gid]
        except KeyError:
            # Entity did not have any components. Nothing to do.
            pass

    def remove_entities_gid(self, gids):
        """ Remove all given entities. """
        # TODO: Write list comprehension for this
        for gid in gids:
            for component in self.entity_components[gid].keys():
                self.component_to_entities[component].remove(gid)
            print ""
        [self.entity_components.pop(gid) for gid in gids]

class ProcessorManager(object):
    """ Keeps track of all processors and notifies them of entity-changes.
        Attributes:
          self.processors      - The list of processors that have been registered with this
                                 instance.
          self.processor_types - The types of the registered processors.
          self.quidditas       - The quidditas instance related to this processor manager.
    """

    def __init__(self, quidditas):
        """ Create an empty ProcessorManager. """
        self.processors = []
        self.processor_types = []
        self.quidditas = quidditas

    def update(self, time):
        """ Perform an update step on all processors that have auto_update enabled. """
        # Disabled auto update
        #[processor.update(time) if processor.auto_update else None for processor in self.processors]
        pass

    def add_processor(self, processor):
        """ Add the processor to the entity system.
        Only add processors that have not been added yet. """
        if processor.__class__ not in self.processor_types:
            self.processor_types.append(processor.__class__)
            self.processors.append(processor)
            return True
        return False

    def remove_processor(self, processor):
        """ Remove the processor from the entity system. """
        if processor.__class__ in self.processor_types:
            self.processor_types.remove(processor.__class__)
            self.processors.remove(processor)

    def refresh_entity(self, entity):
        """ Notify all processors when an entity has changed its components. """
        [processor.refresh_entity(entity) for processor in self.processors]

    def remove_entities_gid(self, gids):
        """ Remove the set of entities from all processors. """
        [processor.remove_entities_gid(gids) for processor in self.processors]

class Processor(object):
    """
    Keeps a list of entities that have relevant components and works with these components
    on each update.
        Attributes:
          self.auto_update - Defaults to true, enables the processor manager to
                             automatically update this processor on every update. """

    def __init__(self):
        """
        Create a new processor that operates on the given component types.
        """
        self.component_manager = None
        self.entities = set()
        try:
            if self.component_types is not None and type(self.component_types) is not set:
                self.component_types = set(self.component_types)
        except AttributeError:
            self.component_types = set()

    def set_component_manager(self, manager):
        self.component_manager = manager
        self.entity_components = manager.entity_components

    def set_quidditas(self, q):
        """ Store a reference to the current quidditas instance that this processor was added to. """
        self.quidditas = q

    def refresh_entities(self, entity_list):
        """ Refresh a list of entities at once. """
        [self.refresh_entity(entity) for entity in entity_list]

    def refresh_entity(self, entity):
        """ Check if the entity has all components
        that are relevant to this processor and add
        it to the list of entities in that case.

        When it doesn't and has been an entity of this processor, remove it.

        Calls added() when the entity is added, removed() when it is removed.
        """
        component_types = entity.get_component_types()
        if self.component_types.issubset(set(component_types)):
            self.entities.add(entity)
            self.added(entity)
        elif entity in self.entities:
            self.entities.remove(entity)
            self.removed(entity)

    def update(self, time):
        """ Update all assigned entities. """
        [self.update_entity(time, entity, self.entity_components[entity.gid]) for entity in self.entities]

    def update_entity(self, time, entity, components):
        """ Work with the components of the given entity.
        Component instances are accessed by components[component_type].
        """
        pass

    def added(self, entity):
        """ Called when an entity of interest was added to the world. """
        pass

    def removed(self, entity):
        """ Called when an entity of interest was removed from the world. """
        pass

    def remove_entities_gid(self, gids):
        """ Remove all given entity ids that might be associated with this processor. """
        remove = set([e if e.gid in gids else '' for e in self.entities])
        self.entities = self.entities.difference(remove)

class Entity(object):
    """ A single entity. This is usually never subclassed.
    The id is set by the entity manager.
    """

    def __init__(self, type_name=None):
        """
        Create a new entity with an optional type.
        """
        self.gid = 0
        self.quidditas = None

    def kill(self):
        """ Remove this entity from the current quidditas. """
        self.quidditas.remove_entity_gid(self.gid)

    def add_component(self, component):
        """ Add a component to this entity. """
        try:
            self.quidditas.add_component(self, component)
        except AttributeError:
            # Entity has not yet been added to a system, cache the components for later addition.
            try:
                self.components.append(component)
            except AttributeError:
                self.components = [component,]

    def get_type(self):
        """ Return the type name of this entity or None when no type was defined. """
        try:
            return self.get_component(TypeComponent).type
        except NameError:
            return None

    def remove_component(self, component):
        """ Remove a component from this entity. """
        try:
            self.quidditas.remove_component(self, component)
        except AttributeError:
            # Try removing the component from the cached ones.
            try:
                self.components.remove(component)
            except:
                # This entity does not have any components cached, nothing to remove then.
                return

    def get_component(self, component_type):
        """ Retrieve the component instance for the given component type that is associated with this entity. """
        try:
            return self.quidditas.component_manager.entity_components[self.gid][component_type]
        except KeyError:
            raise LookupError("No component of type " + str(component_type))
        except AttributeError:
            try:
                for component in self.components:
                    if component.__class__ == component_type:
                        return component
            except AttributeError:
                # No component of this type associated with this.s
                raise LookupError("No component of type " + str(component_type))

    def get_component_types(self):
        """ Return the types of all components associated with this entity.
        When the entity has not beed added to a quidditas instance and has no local
        components, an empty list will be returned. """
        try:
            return self.quidditas.component_manager.entity_components[self.gid].keys()
        except KeyError:
            # No Components associated with this entity.
            return ()
        except AttributeError:
            try:
                comps = []
                for comp in self.components:
                    comps.append(comp.__class__)
                return comps
            except AttributeError:
                return ()

class TypeComponent(object):
    """ A default component for every entity containing the type of the entity. """
    def __init__(self, type_name):
        self.type = type_name



# quidditas setup.py
from distutils.core import setup

setup(
    name = "quidditas",
    packages = ["quidditas"],
    version= "0.9",
    description = "Entity framework for games",
    author = "Dominik Schacht",
    author_email = "domschacht@gmail.com",
    url = "http://quidditas.googlecode.com/",
    keywords = ["entity", "entities", "games"],
    license = "BSD",
    classifiers = [
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Developers",
        "Development Status :: 4 - Beta",
        "Topic :: Games/Entertainment",
    ]

)


